
class BankCard{
  var pin : int;
  var accNo : int;
  var valid : bool;
  

  method Init(p : int, a : int)
  modifies this;
  ensures valid == true;
  ensures pin == p;
  ensures accNo == a;
  {
    pin := p; accNo := a; valid := true;
  }
  
  method isPinOK(p : int ) returns (res: bool)
  ensures res <==> (valid && pin == p);
  {
    if (!this.valid) {res := false;}
    else {res := this.pin == p;}
  }
  
  method invalidateCard()
  modifies this`valid;
  ensures !valid;
  {
    valid := false;
  }
  
  method isCardValid() returns (res: bool)
  ensures res == this.valid;
{
    res := this.valid;
  }
  
  method getAccNo() returns (res : int)
  ensures res == this.accNo;
  {
    res := this.accNo;
  }

function getPin() : int
   reads this;
  {
    this.pin
  }
  
function isValid() : bool
         reads this; {this.valid}
}



class ATM{
  var insertedCard : BankCard;
  var wrongPINCounter : int;
  var auth : bool;
  
  predicate ready()
  reads this;
  {insertedCard == null && wrongPINCounter == 0 && auth == false}

  method init()
  modifies this;
  ensures wrongPINCounter == 0;
  ensures auth == false;
  ensures insertedCard == null;
  {
  insertedCard := null; wrongPINCounter := 0; auth := false;
  }

  method insertCard(c : BankCard)
   modifies this`insertedCard;
   requires c != null && c.valid;
   requires this.ready();
   ensures insertedCard == c;
  {
  insertedCard := c;
  }

  method spitCardOut() returns (card : BankCard)
   modifies this;
   requires insertedCard != null;
   ensures card == old(insertedCard);
   ensures insertedCard == null;
   ensures wrongPINCounter == 0;
   ensures auth == false;       
  {
  card := insertedCard;
  insertedCard := null; 
  wrongPINCounter := 0; 
  auth := false; 
  }

  method enterPIN(p : int)
  // This method may modify the ATM object, and its BankCard field.
  modifies this`auth, this`wrongPINCounter, this.insertedCard`valid;

  // In order for this method to be applicable, a card must have been inserted,
  // the card is not yet authenticated and the card must be valid.
  requires insertedCard != null;
  requires !auth;
  requires insertedCard.valid;
  
  // These caluses specify what DO change in the different scenarios.
  ensures p == insertedCard.getPin() ==> 
  auth && insertedCard.valid == old(insertedCard.valid) && wrongPINCounter == old(wrongPINCounter);

  ensures (p != insertedCard.getPin() && old(wrongPINCounter) < 2) ==> 
    !auth && wrongPINCounter == old(wrongPINCounter) + 1 && insertedCard.valid == old(insertedCard.valid);

  ensures (p != insertedCard.getPin() && old(wrongPINCounter) >= 2) ==> 
  !auth && !insertedCard.valid && wrongPINCounter == old(wrongPINCounter); 
  {
    if (p==insertedCard.pin){
      auth := true;
    }
    else{
      if(wrongPINCounter < 2)
        {wrongPINCounter := wrongPINCounter+1;}
      else
        {insertedCard.invalidateCard();}
    }
  }

method Main(){
  var atm := new ATM;
  atm.init();

  var card := new BankCard;
  card.Init(1234, 987654);
  atm.insertCard(card);

  assert atm.wrongPINCounter == 0; 

  atm.enterPIN(5678); 
  assert !atm.auth;
  assert atm.insertedCard.valid;
  assert atm.wrongPINCounter == 1;

  atm.enterPIN(5679); 
  assert !atm.auth;
  assert atm.insertedCard.valid;
  assert atm.wrongPINCounter == 2;

  //atm.enterPIN(5670); 
  //assert !atm.auth;
  //assert !atm.insertedCard.valid;

  atm.enterPIN(1234);
  assert atm.auth;
  

}

}